/*
 * Copyright (C) 2011-2023 CERN (www.cern.ch)
 * Author: Aurelio Colosimo
 * Based on PTPd project v. 2.1.0 (see AUTHORS for details)
 *
 * Released according to the GNU LGPL, version 2.1 or any later version.
 */

#include <ppsi/ppsi.h>
#include <ppsi/bmc_cmp.h>

char *bmc_cmp_reason[] = {
	[PPSI_BMC_CMP_REASON_UNKNOWN] = 	"unknown",
	[PPSI_BMC_CMP_REASON_NOT_QUALIFIED] =	"not qualified",
	[PPSI_BMC_CMP_REASON_PRIO1] =		"prio1",
	[PPSI_BMC_CMP_REASON_PRIO2] =		"prio2",
	[PPSI_BMC_CMP_REASON_CLOCKCLASS] =	"clock class",
	[PPSI_BMC_CMP_REASON_CLOCKACCURACY] =	"clock accuracy",
	[PPSI_BMC_CMP_REASON_LOGVARIANCE] =	"log variance",
	[PPSI_BMC_CMP_REASON_CLOCKID] =		"clock id",
	[PPSI_BMC_CMP_REASON_PORTNUMBER] =	"port number",
	[PPSI_BMC_CMP_REASON_STEPSREMOVED] =	"steps removed",
	[PPSI_BMC_CMP_REASON_SOURCEPORTID] =	"peer portId",
	[PPSI_BMC_CMP_REASON_RECEIVEPORTID] =	"port portId",
};

int bmc_idcmp(struct ClockIdentity *a, struct ClockIdentity *b)
{
	return memcmp(a, b, sizeof(*a));
}

int bmc_idcmp_reason(struct ClockIdentity *a, struct ClockIdentity *b,
		     int *reason)
{
	int ret;

	ret = memcmp(a, b, sizeof(*a));
	if (ret)
		*reason = PPSI_BMC_CMP_REASON_CLOCKID;

	return ret;
}

int bmc_pidcmp(struct PortIdentity *a, struct PortIdentity *b)
{
	int reason;

	return bmc_pidcmp_reason(a, b, &reason);
}

int bmc_pidcmp_reason(struct PortIdentity *a, struct PortIdentity *b, int *reason)
{
	int ret;

	ret = bmc_idcmp(&a->clockIdentity, &b->clockIdentity);
	if (ret != 0) {
		* reason = PPSI_BMC_CMP_REASON_CLOCKID;
		return ret;
	}

	*reason = PPSI_BMC_CMP_REASON_PORTNUMBER;

	return a->portNumber - b->portNumber;
}

static int are_qualified(struct pp_instance *ppi,
			 struct pp_frgn_master *a,
			 struct pp_frgn_master *b, int *ret) {
	int a_is_qualified = is_qualified(ppi,a);
	int b_is_qualified = is_qualified(ppi,b);


	/* if B is not qualified  9.3.2.5 c) & 9.3.2.3 a) & b)*/
	if (a_is_qualified && !b_is_qualified) {
		pp_diag(ppi, bmc, 2, "Dataset %s not qualified\n", "B");
		*ret=-1;
		return 0;
	}

	/* if A is not qualified  9.3.2.5 c) & 9.3.2.3 a) & b) */
	if (b_is_qualified && !a_is_qualified) {
		pp_diag(ppi, bmc, 2, "Dataset %s not qualified\n", "A");
		*ret= 1;
		return 0;
	}

	/* if both are not qualified  9.3.2.5 c) & 9.3.2.3 a) & b) */
	if (!a_is_qualified && !b_is_qualified) {
		pp_diag(ppi, bmc, 2, "Dataset %s not qualified\n", "A & B");
		*ret= 0;
		return 0;
	}
	return 1;

}

/* compare part2 of the datasets which is the topology, fig 27, page 89 */
int bmc_gm_cmp(struct pp_instance *ppi,
	       struct pp_frgn_master *a,
	       struct pp_frgn_master *b)
{
	int reason;

	return bmc_gm_cmp_reason(ppi, a, b, &reason);
}

int bmc_gm_cmp_reason(struct pp_instance *ppi,
		      struct pp_frgn_master *a,
		      struct pp_frgn_master *b,
		      int *reason)
{
	int ret;
	struct ClockQuality *qa = &a->grandmasterClockQuality;
	struct ClockQuality *qb = &b->grandmasterClockQuality;
	char clkid_str[26];

	/* bmc_gm_cmp is called several times, so report only at level 2 */
	pp_diag(ppi, bmc, 2, "%s\n", __func__);

	if (!are_qualified(ppi,a,b,&ret)) {
		*reason = PPSI_BMC_CMP_REASON_NOT_QUALIFIED;
		return ret;
	}

	if (a->grandmasterPriority1 != b->grandmasterPriority1) {
		/* "Priority1 A: %i, Priority1 B: %i\n", */
		pp_diag(ppi, bmc, 3, "%s A: %i, %s B: %i\n",
			"Priority1", a->grandmasterPriority1,
			"Priority1", b->grandmasterPriority1);
		*reason = PPSI_BMC_CMP_REASON_PRIO1;
		return a->grandmasterPriority1 - b->grandmasterPriority1;
	}

	if (qa->clockClass != qb->clockClass) {
		/* "ClockClass A: %i, ClockClass B: %i\n", */
		pp_diag(ppi, bmc, 3, "%s A: %i, %s B: %i\n",
			"ClockClass", qa->clockClass,
			"ClockClass", qb->clockClass);
		*reason = PPSI_BMC_CMP_REASON_CLOCKCLASS;
		return qa->clockClass - qb->clockClass;
	}

	if (qa->clockAccuracy != qb->clockAccuracy) {
		/* "ClockAccuracy A: %i, ClockAccuracy B: %i\n", */
		pp_diag(ppi, bmc, 3, "%s A: %i, %s B: %i\n",
			"ClockAccuracy", qa->clockAccuracy,
			"ClockAccuracy", qb->clockAccuracy);
		*reason = PPSI_BMC_CMP_REASON_CLOCKACCURACY;
		return qa->clockAccuracy - qb->clockAccuracy;
	}

	if (qa->offsetScaledLogVariance != qb->offsetScaledLogVariance) {
		/* "Variance A: %i, Variance B: %i\n", */
		pp_diag(ppi, bmc, 3, "%s A: %i, %s B: %i\n",
			"Variance", qa->offsetScaledLogVariance,
			"Variance", qb->offsetScaledLogVariance);
		*reason = PPSI_BMC_CMP_REASON_LOGVARIANCE;
		return qa->offsetScaledLogVariance
				- qb->offsetScaledLogVariance;
	}

	if (a->grandmasterPriority2 != b->grandmasterPriority2) {
		/* "Priority2 A: %i, Priority2 B: %i\n", */
		pp_diag(ppi, bmc, 3, "%s A: %i, %s B: %i\n",
			"Priority2", a->grandmasterPriority2,
			"Priority2", b->grandmasterPriority2);
		*reason = PPSI_BMC_CMP_REASON_PRIO2;
		return a->grandmasterPriority2 - b->grandmasterPriority2;
	}

	pp_diag(ppi, bmc, 3, "GmId A: %s\n",
		format_hex8(clkid_str, a->grandmasterIdentity.id));

	pp_diag(ppi, bmc, 3, "GmId B: %s\n",
		format_hex8(clkid_str, b->grandmasterIdentity.id));

	return bmc_idcmp_reason(&a->grandmasterIdentity,
				&b->grandmasterIdentity,
				reason);
}

/* Compare part2 of the datasets which is the topology.
 * IEEE1588-2019: fig 35, page 162 */
int bmc_topology_cmp(struct pp_instance *ppi,
		     struct pp_frgn_master *a,
		     struct pp_frgn_master *b)
{
	int reason;

	return bmc_topology_cmp_reason(ppi, a, b, &reason);
}

/* Compare part2 of the datasets which is the topology.
 * IEEE1588-2019: fig 35, page 162,
 * IEEE1588a-2023: fig P1.2, page 37 (copy of the above with more detailed
 * references)
 *
 * return values:
 *  0 -- error
 * -2 -- a better
 * -1 -- a better by topology
 *  1 -- b better by topology
 *  2 -- b better
 */
int bmc_topology_cmp_reason(struct pp_instance *ppi,
			    struct pp_frgn_master *a,
			    struct pp_frgn_master *b,
			    int *reason)
{
	int ret;
	struct PortIdentity *pidtxa = &a->sourcePortIdentity;
	struct PortIdentity *pidtxb = &b->sourcePortIdentity;
	struct PortIdentity *pidrxa = &a->receivePortIdentity;
	struct PortIdentity *pidrxb = &b->receivePortIdentity;
	int diff;
	char clkida_str[26];
	char clkidb_str[26];

	/* bmc_topology_cmp is called several times, so report only at level 2
	 */
	pp_diag(ppi, bmc, 2, "%s\n", __func__);

	if (!are_qualified(ppi,a,b,&ret)) {
		return ret;
	}

	diff = a->stepsRemoved - b->stepsRemoved;
	/* Comparison 2-1, Figure P1.2, IEEE1588a-2023 */
	if (diff > 1 || diff < -1) {
		pp_diag(ppi, bmc, 3, "StepsRemoved A: %i, StepsRemoved B: %i, diffx %d\n",
			a->stepsRemoved, b->stepsRemoved, diff);
		*reason = PPSI_BMC_CMP_REASON_STEPSREMOVED;
		/* Return "better" B, #Z (diff > 1)
		 * or "better" A, #Y (diff < -1) */
		return diff > 1 ? 2 : -2;
	}

	/* Comparison 2-2 (left arrow) Figure P1.2, IEEE1588a-2023 */
	if (diff > 0) {
		/* Comparison 2-3, Figure P1.2, IEEE1588a-2023 */
		if (!bmc_pidcmp(pidtxa, pidrxa)) {
			/* Error #E (left) */
			pp_diag(ppi, bmc, 1, "%s:%i: Error 1\n",
				__func__, __LINE__);
			return 0;
		}
		/* #Q, B better by topology than A */
		pp_diag(ppi, bmc, 3, "StepsRemoved A: %i, StepsRemoved B: %i, diffy %d\n",
			a->stepsRemoved, b->stepsRemoved, diff);
		*reason = PPSI_BMC_CMP_REASON_STEPSREMOVED;
		return 1;
	}

	/* Comparison 2-2 (right arrow), Figure P1.2, IEEE1588a-2023 */
	if (diff < 0) {
		if (!bmc_pidcmp(pidtxb, pidrxb)) {
			/* Error #E (right) */
			pp_diag(ppi, bmc, 1, "%s:%i: Error 1\n",
				__func__, __LINE__);
			return 0;
		}
		/* #X, A better by topology than B */
		pp_diag(ppi, bmc, 3, "StepsRemoved A: %i, StepsRemoved B: %i, diffz %d\n",
			a->stepsRemoved, b->stepsRemoved, diff);
		*reason = PPSI_BMC_CMP_REASON_STEPSREMOVED;
		return -1;
	}

	/* Proceed from 2-2 (bottom arrow) Figure P1.2, IEEE1588a-2023 */

	/* stepsRemoved is equal, compare sourcePortIdentity identities
	 * Comparison 2-5, Figure P1.2, IEEE1588a-2023 */
	diff = bmc_pidcmp(pidtxa, pidtxb);
	if (diff) {
		pp_diag(ppi, bmc, 3, "%sId A: %s.%04x, %sId B: %s.%04x, diffaa %d\n",
			"Tx",
			format_hex8(clkida_str, pidtxa->clockIdentity.id),
			pidtxa->portNumber,
			"Tx",
			format_hex8(clkidb_str, pidtxb->clockIdentity.id),
			pidtxb->portNumber, diff);
		*reason = PPSI_BMC_CMP_REASON_SOURCEPORTID;
		/* Return "better by topology" B, #X (diff > 0)
		 * or "better by topology" A, #Q (diff < 0) */
		return diff > 1 ? 1 : -1;
	}

	/* sourcePortIdentity is equal, compare receive port identites, which
	 * is the last decision maker, which has to be different
	 * Comparison 2-5, Figure P1.2, IEEE1588a-2023 */
	pp_diag(ppi, bmc, 3, "%sId A: %s.%04x, %sId B: %s.%04x\n",
		"Rx",
		format_hex8(clkida_str, pidtxa->clockIdentity.id),
		pidtxa->portNumber,
		"Rx",
		format_hex8(clkidb_str, pidtxb->clockIdentity.id),
		pidtxb->portNumber);
		*reason = PPSI_BMC_CMP_REASON_RECEIVEPORTID;

	diff = bmc_pidcmp(pidrxa, pidrxb);
	/* Return "better by topology" B, #X (> 0)
	 * or "better by topology" A, #Q (< 0)
	 * or error #F (== 0) */
	if (diff > 0)
		return 1;
	if (diff < 0)
		return -1;
	return 0;
}


/*
 * Data set comparison between two foreign masters. Return similar to
 * memcmp().  However, lower values take precedence, so in A-B (like
 * in comparisons,   > 0 means B wins (and < 0 means A wins).
 */
int bmc_dataset_cmp(struct pp_instance *ppi,
		    struct pp_frgn_master *a,
		    struct pp_frgn_master *b)
{
	int reason;

	return bmc_dataset_cmp_reason(ppi, a, b, &reason);
}

/*
 * return values:
 *  0 -- error
 * -2 -- a better
 * -1 -- a better by topology
 *  1 -- b better by topology
 *  2 -- b better
 */
int bmc_dataset_cmp_reason(struct pp_instance *ppi,
			   struct pp_frgn_master *a,
			   struct pp_frgn_master *b,
			   int *reason)
{
	char clkid_str[26];
	int diff;
	/* dataset_cmp is called several times, so report only at level 2 */
	pp_diag(ppi, bmc, 2, "%s\n", __func__);
	pp_diag(ppi, bmc, 3, "portId A: %s\n",
		format_hex8(clkid_str, a->sourcePortIdentity.clockIdentity.id));

	pp_diag(ppi, bmc, 3, "portId B: %s\n",
		format_hex8(clkid_str, b->sourcePortIdentity.clockIdentity.id));

	if (!bmc_idcmp(&a->grandmasterIdentity, &b->grandmasterIdentity)) {
		/* Check topology */
		return bmc_topology_cmp_reason(ppi, a, b, reason);
	} else {
		/* Check grandmasters */
		diff = bmc_gm_cmp_reason(ppi, a, b, reason);
		if (diff < 0)
			return -2; /* a better */
		if (diff > 0)
			return 2; /* b better */
		return 0;
	}
}
