/*
 * Copyright (C) 2022 CERN (www.cern.ch)
 * Author: Adam Wujek
 *
 * Released according to the GNU GPL, version 2 or any later version.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <getopt.h>
#include <unistd.h>
#include <ppsi/ppsi.h>
#include <minipc.h>
// #include <ppsi-wrs.h>


#define PPSI_EXPORT_STRUCTURES
#include "ppsi_exports.h"

#define MINIPC_TIMEOUT 5000

static int verbose = 0;

static struct minipc_ch *ptp_ch;

enum input_arg {
	arg_help = 'z' + 1, /* avoid conflicts with short options */
	arg_alpha,
	arg_autoneg,
	arg_delay_req_interval,
	arg_diags,
	arg_diags_inst,
	arg_extension,
	arg_instance,
	arg_port,
	arg_prio1,
	arg_prio2,
	arg_profile,
	arg_sync_interval,
	arg_tracking,
	arg_verbose,
};
	
static struct option long_opts[] = {
	{"help",	                no_argument,       NULL, arg_help},
	{"alpha",                       required_argument, NULL, arg_alpha},
	{"autonegotiation",             required_argument, NULL, arg_autoneg},
	{"delay-req-interval",          required_argument, NULL, arg_delay_req_interval},
	{"diags",                       required_argument, NULL, arg_diags},
	{"diags-inst",                  required_argument, NULL, arg_diags_inst},
	{"extension",                   required_argument, NULL, arg_extension},
	{"instance",                    required_argument, NULL, arg_instance},
	{"ppi",                         required_argument, NULL, arg_instance},
	{"port",                        required_argument, NULL, arg_port},
	{"prio1",                       required_argument, NULL, arg_prio1},
	{"priority1",                   required_argument, NULL, arg_prio1},
	{"prio2",                       required_argument, NULL, arg_prio2},
	{"priority2",                   required_argument, NULL, arg_prio2},
	{"profile",                     required_argument, NULL, arg_profile},
	{"sync-interval",               required_argument, NULL, arg_sync_interval},
	{"tracking",                    required_argument, NULL, arg_tracking},
	{"verbose",                     optional_argument, NULL, arg_verbose},
	{NULL, 0, NULL, 0}
};

void help(char *prgname)
{
	fprintf(stderr, "%s: Use: %s [-v] [-h] [option]\n",
		prgname, prgname);
	fprintf(stderr,
		"The program has the following options:\n"
		"  -h|--help         - print help\n"
		"  -v|--verbose      - verbose output\n"
		"Global parameters:\n"
		"  --alpha=<alpha64bit>\n"
		"                    - change currently used alpha\n"
		"                      the value is in fpa format, like displayed in wr_mon\n"
		"                      NOTE: the changed value is lost on link down/up\n"
		"  --diags=<num>     - change PPSI's global diagnostics to <num>\n"
		"                      order: FSM, Time, Frames, Servo, BMC, Extension, Configuration\n"
		"  --prio1=<num>\n"
		"  --priority1=<num> - set priority 1 to <num>\n"
		"  --prio2=<num>\n"
		"  --priority2=<num> - set priority 2 to <num>\n"
		"  --tracking=<on|off|enable|disable|1|0>\n"
		"                    - enable/disable tracking in servo\n"
		"\n"
		"Parameters specific for PPSi instance:\n"
		"  --instance=<num|all>\n"
		"  --ppi=<num|all>   - Select PPSi instance <num> or all instances\n"
		"  --port=<num|all>  - Select all instances on a defined port <num> or all\n"
		"                      instances\n"
		"  --autonegotiation=<on|off|enable|disable|1|0>\n"
		"                      enable or disable extension autonegotiation\n"
		"  --delay-req-interval=<num>\n"
		"                    - sets logarithm to the base 2 of the mean interval of delay\n"
		"                      request message transmission; used when a port is in Slave state\n"
		"  --diags-inst=<num>\n"
		"                    - change PPSI's instance/port diagnostics to <num>\n"
		"                      order: FSM, Time, Frames, Servo, BMC, Extension, Configuration\n"
		"  --extension=<none|wr|l1s|l1sync>\n"
		"                    - sets the extension; extension has to be supported by a selected profile\n"
		"  --profile=<ptp|wr|ha|ha_wr|custom>\n"
		"                    - sets the profile; ha_wr, ha and wr are the same profile\n"
		"                      setting a profile (also to the current one) may change number\n"
		"                      of attributes to its default values (not necessary the same\n"
		"                      as defined in ppsi.conf);\n"
		"                      changing a profile to custom does not change any parameters;\n"
		"  --sync-interval=<num>\n"
		"                    - sets logarithm to the base 2 of the mean interval of sync\n"
		"                      message transmission; used when a port is in Master state\n"
		"\n\n"
		"Version: " __GIT_VER__ " compiled by " __GIT_USR__ " on " __DATE__ ", " __TIME__ "\n"
	);
	exit(1);
}


void ppsi_connect_minipc(void)
{
	if (ptp_ch) {
		/* close minipc, if connected before */
		minipc_close(ptp_ch);
	}
	ptp_ch = minipc_client_create("ptpd", 0);
	if (!ptp_ch) {
		fprintf(stderr, "Can't establish WRIPC connection to the PTP "
			 "daemon!\n");
		exit(1);
	}
}

void init_shm(void)
{
	ppsi_connect_minipc();
}

static void check_instance(int instance)
{
	if (instance == -1) {
		fprintf(stderr, "Instance not defined! Please define instance."
			"Exiting...\n");
		exit (1);
	}
}

static void enable_disable_tracking(int onoff) {

	int rval;
	
	printf("%s: onoff %d\n", __func__, onoff);
	minipc_call(ptp_ch, MINIPC_TIMEOUT, &__rpcdef_cmd, &rval, 1, onoff);
}

static int ppsi_set_param(int param, int val) {

	int rval = 0;
	
	printf("%s: set param %d val %d\n", __func__, param, val);
	minipc_call(ptp_ch, MINIPC_TIMEOUT, &ppsiexp_update_param_cmd, &rval,
		    param, val);

	return rval;
}

static int ppsi_set_param64(int param, int64_t val64) {

	int rval = 0;
	
	printf("%s: set param %d val %lld\n", __func__, param, val64);
	minipc_call(ptp_ch, MINIPC_TIMEOUT, &ppsiexp_update_param64_cmd, &rval,
		    param, val64);

	return rval;
}

static int ppsi_set_param_str(int param, char *val)
{
	int rval = 0;

	printf("%s: set param (%d), val %s\n",
	       __func__, param, val);
	minipc_call(ptp_ch, MINIPC_TIMEOUT,
		    &ppsiexp_update_param_str_cmd, &rval, param, val);

	return rval;
}

static int ppsi_set_param_instance(int param, int instance, int val)
{
	int rval = 0;
	
	printf("%s: set param (%d) for %s %d, val %d\n",
	       __func__, param,
	       instance & PPSI_INSTANCE_USE_PORT ? "port" : "instance",
	       instance & ~PPSI_INSTANCE_USE_PORT, val);
	minipc_call(ptp_ch, MINIPC_TIMEOUT, &ppsiexp_update_param_instance_cmd, &rval,
		    param, instance, val);

	return rval;
}

static int ppsi_set_param_instance_str(int param, int instance, char *val)
{
	int rval = 0;

	printf("%s: set param (%d) for %s %d, val %s\n",
	       __func__, param,
	       instance & PPSI_INSTANCE_USE_PORT ? "port" : "instance",
	       instance & ~PPSI_INSTANCE_USE_PORT, val);
	minipc_call(ptp_ch, MINIPC_TIMEOUT,
		    &ppsiexp_update_param_instance_str_cmd, &rval, param,
		    instance, val);

	return rval;
}

int main(int argc, char *argv[])
{
	int opt;
	int tmp;
	char *tmp_str;
	int ret;
	int ppsi_instance = -1;
	int ppsi_port = -1;

	/* If no params print help */
	if (argc == 1)
		help(argv[0]);

	while ((opt = getopt_long(argc, argv, "vh", long_opts, NULL)) != -1) {
		if (opt == 'h' || opt == arg_help )
			help(argv[0]);
		if (opt == '?') {
			/* Unrecognized option, exit */
			exit(1);
		}

		if (opt == 'v') {
			verbose = 1;
			printf("Enable verbose mode\n");
		}

		if (opt == arg_verbose) {
			if (!strcmp(optarg, "off") || !strcmp(optarg, "0")) {
				verbose = 0;
				printf("Enable verbose mode\n");
			} else {
				verbose = 1;
				printf("Disable verbose mode\n");
			}
		}
	}

	init_shm();

	/* Reset position of argument for getopt_long */
	optind = 0;

	while ((opt = getopt_long(argc, argv, "vh", long_opts, NULL)) != -1) {

		switch (opt) {
		case arg_alpha:
			if (verbose)
				printf("Setting alpha to %lld\n",
				       atoll(optarg));
			ret = ppsi_set_param64(PPSIEXP_PARAM64_ALPHA_CMD,
					       atoll(optarg));
			if (ret) {
				fprintf(stderr, "Error setting alpha to %lld\n",
					atoll(optarg));
				exit (1);
			}
			break;

		case arg_autoneg:
			/* Check if an instance is set. If not exit. */
			check_instance(ppsi_instance);

			if (!strcmp(optarg, "on")
			    || !strcmp(optarg, "enable")
			    || !strcmp(optarg, "1")
			    ) {
				tmp = PPSI_EXT_AUTONEG_ENABLE;
			} else if (!strcmp(optarg, "off")
				    || !strcmp(optarg, "disable")
				    || !strcmp(optarg, "0")
				    ) {
				tmp = PPSI_EXT_AUTONEG_DISABLE;
			} else {
				fprintf(stderr, "Unknown value (%s) for"
					" autonegotiation\n", optarg);
				exit (1);
			}

			if (verbose)
				printf("Setting autonegotiation for "
				       "%s %d to %s\n",
				       ppsi_port == -1 ? "instance" : "port",
				       ppsi_instance & ~PPSI_INSTANCE_USE_PORT,
				       tmp ? "enable" : "disable");
			ret = ppsi_set_param_instance(
					PPSIEXP_PARAM_INST_AUTONEG_CMD,
					ppsi_instance,
					tmp);
			if (ret) {
				fprintf(stderr, "Error setting autonegotiation "
					"for %s %d to %s, ret %d\n",
					ppsi_port == -1 ? "instance" : "port",
					ppsi_instance & ~PPSI_INSTANCE_USE_PORT,
					tmp ? "enable" : "disable", ret);
				exit (1);
			}
			break;

		case arg_delay_req_interval:
			/* Check if an instance is set. If not exit. */
			check_instance(ppsi_instance);
			tmp = atoi(optarg);
			if (verbose)
				printf("Setting delay request interval for "
				       "%s %d to %d\n",
				       ppsi_port == -1 ? "instance" : "port",
				       ppsi_instance & ~PPSI_INSTANCE_USE_PORT,
				       tmp);
			ret = ppsi_set_param_instance(
					PPSIEXP_PARAM_INST_DELAY_REQ_INT_CMD,
					ppsi_instance,
					tmp);
			if (ret) {
				fprintf(stderr, "Error setting delay request "
					"interval for %s %d to %d, ret %d\n",
					ppsi_port == -1 ? "instance" : "port",
					ppsi_instance & ~PPSI_INSTANCE_USE_PORT,
					tmp, ret);
				exit (1);
			}
			break;

		case arg_diags:
			if (verbose)
				printf("Setting global diagnostics to %s\n",
				       optarg);
			ret = ppsi_set_param_str(PPSIEXP_PARAM_STR_DIAG_CMD,
						 optarg);
			if (ret) {
				fprintf(stderr, "Error setting "
					"global diagnostics to %s, ret %d\n",
					optarg, ret);
				exit (1);
			}
			break;

		case arg_diags_inst:
			/* Check if an instance is set. If not exit. */
			check_instance(ppsi_instance);
			if (verbose)
				printf("Setting diagnostics for "
				       "%s %d to %s\n",
				       ppsi_port == -1 ? "instance" : "port",
				       ppsi_instance & ~PPSI_INSTANCE_USE_PORT,
				       optarg);
			ret = ppsi_set_param_instance_str(
					PPSIEXP_PARAM_INST_STR_DIAG_CMD,
					ppsi_instance,
					optarg);
			if (ret) {
				fprintf(stderr, "Error setting delay request "
					"interval for %s %d to %s, ret %d\n",
					ppsi_port == -1 ? "instance" : "port",
					ppsi_instance & ~PPSI_INSTANCE_USE_PORT,
					optarg, ret);
				exit (1);
			}
			break;

		case arg_extension:
			/* Check if an instance is set. If not exit. */
			check_instance(ppsi_instance);
			if (!strcmp(optarg, "none") || !strcmp(optarg, "0")) {
				tmp = PPSIEXP_PARAM_EXTENSION_NONE;
				tmp_str = "none";
			} else if (!strcmp(optarg, "wr") || !strcmp(optarg, "1")) {
				tmp = PPSIEXP_PARAM_EXTENSION_WR;
				tmp_str = "wr";
			} else if (!strcmp(optarg, "l1s")
			    || !strcmp(optarg, "l1sync")
			    || !strcmp(optarg, "2")) {
				tmp = PPSIEXP_PARAM_EXTENSION_L1SYNC;
				tmp_str = "l1sync";
			} else {
				fprintf(stderr, "Error setting extension for %s"
					" %d to %s, wrong value\n",
					ppsi_port == -1 ? "instance" : "port",
					ppsi_instance & ~PPSI_INSTANCE_USE_PORT,
					optarg);
				exit (1);
			}

			if (verbose)
				printf("Setting extension for %s %d to %s\n",
				       ppsi_port == -1 ? "instance" : "port",
				       ppsi_instance & ~PPSI_INSTANCE_USE_PORT,
				       tmp_str);
				ret = ppsi_set_param_instance(
					PPSIEXP_PARAM_INST_EXTENSION_CMD,
					ppsi_instance,
					tmp);
			if (ret) {
				fprintf(stderr, "Error setting extension for %s"
					" %d to %s, ret %d\n",
					ppsi_port == -1 ? "instance" : "port",
					ppsi_instance & ~PPSI_INSTANCE_USE_PORT,
					tmp_str, ret);
				exit (1);
			}
			break;

		case arg_instance:
			ppsi_instance = atoi(optarg);
			if (ppsi_instance == 0 && strcmp("0", optarg)
			    && strcmp("all", optarg)) {
				fprintf(stderr, "Wrong instance value: %s\n",
					optarg);
				exit (1);
			}
			if (!strcmp("all", optarg)) {
				ppsi_port = 0;
				ppsi_instance = PPSI_INSTANCE_USE_PORT;
			} else {
				ppsi_port = -1;
			}
			if (verbose)
				printf("Setting instance to %d%s\n",
				       ppsi_instance & ~PPSI_INSTANCE_USE_PORT,
				       ppsi_instance == PPSI_INSTANCE_USE_PORT
					    ? " (all)" : ""
				      );

			break;

		case arg_port:
			ppsi_instance = atoi(optarg);
			if (ppsi_instance == 0 && strcmp("all", optarg)) {
				fprintf(stderr, "Wrong port value: %s\n",
					optarg);
				exit (1);
			}
			ppsi_port = ppsi_instance;
			ppsi_instance |= PPSI_INSTANCE_USE_PORT;
			if (verbose)
				printf("Setting port to %d%s\n",
				       ppsi_port,
				       ppsi_instance == PPSI_INSTANCE_USE_PORT
					    ? " (all)" : ""
				      );
			break;

		case arg_prio1:
			tmp = atoi(optarg);
			if (verbose)
				printf("Setting priority1 to %d\n", tmp);
			ret = ppsi_set_param(PPSIEXP_PARAM_PRIORITY1_CMD, tmp);
			if (ret) {
				fprintf(stderr, "Error setting priority1 to "
					"%d\n", tmp);
				exit (1);
			}
			break;

		case arg_prio2:
			tmp = atoi(optarg);
			if (verbose)
				printf("Setting priority2 to %d\n", tmp);
			ret = ppsi_set_param(PPSIEXP_PARAM_PRIORITY2_CMD, tmp);
			if (ret) {
				fprintf(stderr, "Error setting priority2 to "
					"%d\n", tmp);
				exit (1);
			}
			break;

		case arg_profile:
			tmp = -1;
			/* Check if an instance is set. If not exit. */
			check_instance(ppsi_instance);
			if (!strcmp(optarg, "ptp")) {
				tmp = PPSIEXP_PARAM_PROFILE_PTP;
				tmp_str = "ptp";
			}
			if (!strcmp(optarg, "ha_wr")
			    || !strcmp(optarg, "ha")
			    || !strcmp(optarg, "wr")) {
				tmp = PPSIEXP_PARAM_PROFILE_HA_WR;
				tmp_str = "ha_wr";
			}
			if (!strcmp(optarg, "custom")) {
				tmp = PPSIEXP_PARAM_PROFILE_CUSTOM;
				tmp_str = "custom";
			}

			if (tmp == -1) {
				fprintf(stderr, "Wrong Value! unable to set "
					"global profile to %s\n", optarg);
				ret = PPSIEXP_RET_ERROR_VAL;
				break;
			}

			if (verbose)
				printf("Setting profile for %s %d to %s\n",
				       ppsi_port == -1 ? "instance" : "port",
				       ppsi_instance & ~PPSI_INSTANCE_USE_PORT,
				       tmp_str);
				ret = ppsi_set_param_instance(
					PPSIEXP_PARAM_INST_PROFILE_CMD,
					ppsi_instance,
					tmp);
			if (ret) {
				fprintf(stderr, "Error setting profile for %s "
					"%d to %s, ret %d\n",
					ppsi_port == -1 ? "instance" : "port",
					ppsi_instance & ~PPSI_INSTANCE_USE_PORT,
					tmp_str, ret);
				exit (1);
			}
			break;

		case arg_sync_interval:
			/* Check if an instance is set. If not exit. */
			check_instance(ppsi_instance);
			tmp = atoi(optarg);
			if (verbose)
				printf("Setting sync interval for %s %d"
				       "to %d\n",
				       ppsi_port == -1 ? "instance" : "port",
				       ppsi_instance & ~PPSI_INSTANCE_USE_PORT,
				       tmp);
				ret = ppsi_set_param_instance(
					PPSIEXP_PARAM_INST_SYNC_INT_CMD,
					ppsi_instance,
					tmp);
			if (ret) {
				fprintf(stderr, "Error setting sync interval "
					"for %s %d to %d, ret %d\n",
					ppsi_port == -1 ? "instance" : "port",
					ppsi_instance & ~PPSI_INSTANCE_USE_PORT,
					tmp, ret);
				exit (1);
			}
			break;

		case arg_tracking:
			{
				int tracking;
				if (!strcmp(optarg, "on")
				    || !strcmp(optarg, "enable")
				    || !strcmp(optarg, "1")
				   ) {
					tracking = 1;
				} else if (!strcmp(optarg, "off")
					   || !strcmp(optarg, "disable")
					   || !strcmp(optarg, "0")
					  ) {
					tracking = 0;
				} else {
					fprintf(stderr, "Unknown value (%s) for"
					        " tracking\n", optarg);
					exit (1);
				}
				if (verbose)
					printf("%s tracking\n",
					       tracking ? "Enable" : "Disable");

				enable_disable_tracking(tracking);
				break;
			}
		default:
			break;
		}
	}

	return 0;
}
